/*
   init.c -- PIKT initialization routines

   $Header: /var/cvs/PIKT/pikt/src/lib/init.c,v 1.54 2004/09/04 04:59:26 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002,2003,2004 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
*/

#include "common.h"

#include <netdb.h>

FILE *confin;

int ismaster = FALSE;

char hostname[MAXIDENT+1];
char domainname[MAXIDENT+1];

/* PIKT.conf options */
AUTHUNIX_CREATE_UID_T authuid = 0;		/* default to root */
AUTHUNIX_CREATE_GID_T authgid = 0;		/* default to root */
char *master = (char *) NULL;
char *domain = (char *) NULL;
char *accessmode = (char *) NULL;
char *master_address = (char *) NULL;
char *master_ethernet_address = (char *) NULL;
int data_encryption_type = 0;			/* default to no encryption
						   of data
						 */
int auth_encryption_type = 0;			/* default to no encryption
						   of key
						 */
char *pvtkey = (char *) NULL;
unsigned int pvtkeylen = 0;
int isauth_by_master = TRUE;
int isauth_by_master_address = FALSE;
int isauth_by_master_ethernet_address = FALSE;
int rpc_timeout = RPC_TIMEOUT;
int rpc_retry_timeout = RPC_RETRY_TIMEOUT;
int rpcinfo_retries = RPCINFO_RETRIES;
int restart_piktc_svc_wait = RESTART_PIKTC_SVC_WAIT;
int isverboselog = FALSE;
int isautosyslog = FALSE;
int istcponly = TRUE;
int execcmdout_limit = EXECCMDOUT_LIMIT;

int PIKT_conf_gid = FALSE;

char *homdir = (char *) NULL;	/* be explicit about initialization, just
				   to be sure (also for all that follow)
				 */
char *bindir = (char *) NULL;
char *etcdir = (char *) NULL;
char *lokdir = (char *) NULL;
char *libdir = (char *) NULL;
char *vardir = (char *) NULL;
char *altdir = (char *) NULL;
char *prgdir = (char *) NULL;
char *objdir = (char *) NULL;
char *cfgdir = (char *) NULL;
char *difdir = (char *) NULL;
char *stgdir = (char *) NULL;
char *hstdir = (char *) NULL;
char *logdir = (char *) NULL;
char *PIKTconf = (char *) NULL;
char *keysconf = (char *) NULL;
char *pikt = (char *) NULL;
char *piktd = (char *) NULL;
char *piktc = (char *) NULL;
char *piktc_svc = (char *) NULL;
char *pikts = (char *) NULL;
char *piktx = (char *) NULL;
char *piktdconf = (char *) NULL;
char *piktdconftmp = (char *) NULL;
char *piktdlock = (char *) NULL;
char *piktdlog = (char *) NULL;
char *piktdalt = (char *) NULL;
char *piktclock = (char *) NULL;
char *piktclog = (char *) NULL;
char *piktc_svclock = (char *) NULL;
char *piktc_svclog = (char *) NULL;

char *arpcmd = (char *) NULL;
char *cksumcmd = (char *) NULL;
char *cpcmd = (char *) NULL;
char *diffcmd = (char *) NULL;
char *echocmd = (char *) NULL;
char *egrepcmd = (char *) NULL;
char *filecmd = (char *) NULL;
char *pscmd = (char *) NULL;
char *rmcmd = (char *) NULL;
char *nicecmd = (char *) NULL;
char *catcmd = (char *) NULL;
char *rpcinfocmd = (char *) NULL;
char *scpcmd = (char *) NULL;
char *sumcmd = (char *) NULL;
char *tailcmd = (char *) NULL;
char *trcmd = (char *) NULL;

int echocmdlen;

char *bakext = (char *) NULL;

unsigned long confparsrv;
char **confparstrp;
int *confparintp;

int endian;

time_t time_start;
pid_t prgpid;
pid_t prgppid;
char *prgnam;

int exstat = 0;
const int exit_status[] = { 2,			/* LOG_EMERG   */
                            2,			/* LOG_ALERT   */
                            2,			/* LOG_CRIT    */
                            2,			/* LOG_ERR     */
                            1,			/* LOG_WARNING */
                            0,			/* LOG_NOTICE  */
                            0,			/* LOG_INFO    */
                            0			/* LOG_DEBUG   */
                          } ;
const char *error_level[] = { "ERROR",		/* LOG_EMERG   */
                              "ERROR",		/* LOG_ALERT   */
                              "ERROR",		/* LOG_CRIT    */
                              "ERROR",		/* LOG_ERR     */
                              "WARNING",	/* LOG_WARNING */
                              "INFO",		/* LOG_NOTICE  */
                              "INFO",		/* LOG_INFO    */
                              "DEBUG"		/* LOG_DEBUG   */
                            };

/* up to MAXIF (64) tabs */
const char *tabs[] = {
                       "",
                       "\t",
                       "\t\t",
                       "\t\t\t",
                       "\t\t\t\t",
                       "\t\t\t\t\t",
                       "\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                       "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
                     };

int isopt(char **, char **, int);
void gethome(char **);
void initpaths(char *);
void getkey(char **, char *, char *);

/* perform various initializations
 */
void
init (int argc, char **argv, void (*usage)(void), const char *verstr)
{
	static char fnnam[] = "init";
	char *cp;
	char *fqdn = (char *)NULL;
	struct hostent *hp;
	char *hap;

	istest = isopt(argv+1, argv+argc-1, 'T');
	isdebug = isopt(argv+1, argv+argc-1, 'G');
	if((isshow_version = isopt(argv+1, argv+argc-1, 'V'))) {
		show_version(verstr);
		bailout = TRUE;
	}
	if((ishelp = isopt(argv+1, argv+argc-1, 'h'))) {
		(*usage)();
		bailout = TRUE;
	}
	if (bailout)
		return;

	/* get the fully qualified domain name of this host */
	gethostname(hostname, MAXIDENT);
	hostname[MAXIDENT] = '\0';	/* to protect against if we have not
					   made hostname[] sufficiently large;
					   from the gethostname() man page:
					   "returned name is null-terminated
					   unless insufficient space is
					   provided"
					 */
	cp = strchr(hostname, '.');
	if (! cp) {
		hp = gethostbyname(hostname);
		if (! hp || ! hp->h_name) {
			logexit(OUTPUT, LOG_ERR, ERRID_GETHOSTBYNAME_ERROR, __FILE__, __LINE__, fnnam, FALSE, "for hostname %s, gethostbyname() returns null pointer, or null pointer for h_name.", hostname);
		}
		strncpy(hostname, hp->h_name, minval(MAXIDENT, strlen(hp->h_name)));
		cp = strchr(hostname, '.');
		if (! cp) {
			/* walk the aliases list */
			while ((hap = *(hp->h_aliases)) && ! cp) {
				strncpy(hostname, hap, minval(MAXIDENT, strlen(hap)));
				cp = strchr(hostname, '.');
				hp->h_aliases++;
			}
		}
	}
	if (cp) {	/* we have the fully qualified domain name */
		fqdn = mystrdup(hostname, fnnam);
		if (! istest)
			*cp = '\0';
		/* given that hostname[] and domainname[] are of equal
		   length, by logic the length of the domainname[] buffer >=
		   strlen(cp+1), so there is no risk of buffer overflow with
		   the following strcpy()
		 */
		strcpy(domainname, cp+1);
	} else {	/* else we don't have it, so just give up! */
		logexit(OUTPUT, LOG_ERR, ERRID_FQDN_INDETERMINATE, __FILE__, __LINE__, fnnam, FALSE, "unable to determine the fully qualified domain name of this host.");
	}

	gethome(argv);		/* determine PIKT home dir */

	getconf(homdir);	/* get config stuff from PIKT.conf */

	if (istest) {
		FREE(master);
		master = mystrdup(hostname, fnnam);
		FREE(domain);
		domain = mystrdup("domain", fnnam);
		FREE(accessmode);
		accessmode = mystrdup("local", fnnam);
	}

	if (bailout)
		return;

	initpaths(homdir);	/* initialize dir/file paths */

	if (bailout)
		return;

	/* from here on, it is safe to call either logreturn() or logexit()
	   since we have the paths to the log files
	 */

	getkey(&pvtkey, "private_key", etcdir);		/* get private_key */

	if (bailout)
		return;

	if (fqdn) {
		if (! strcmp(fqdn, master))
			ismaster = TRUE;
		free(fqdn);
	} else {
		if (! strncmp(hostname, master, (size_t)strlen(hostname)))
			ismaster = TRUE;
	}

	sig_init();		/* initialize signal handlers */

	/* get the program process id */
	prgpid = getpid();

	/* get the program parent process id */
	prgppid = getppid();

	/* save the start time */
	time_start = time((time_t*)0);

	/* seed the random number generator with several semi-random
	   elements
	 */
	srand((unsigned)(authuid+authgid+prgpid+prgppid+time_start));

	/* set this machine's endianness */
	endian =
	#ifdef WORDS_BIGENDIAN
	         BIGENDIAN;
	#else
	         LITTLEENDIAN;
	#endif

	return;
}

/* look for a command-line (minus) switch, returning TRUE if found, else
   FALSE; this is for simpler command-line parsing; for more involved parsing,
   we use cmd.[lyc] and cmdparse()
 */
int
isopt (char **avcurr, char **avlast, int opt)
{
	while (avcurr <= avlast) {
		if (**avcurr == '-') {
			if (strchr(*avcurr, opt))
				return TRUE;
		}
		avcurr++;
	}
	return FALSE;
}

/* determine the PIKT home directory, either directly from argv (when
   the program is invoked with a full path), or by searching through the
   environment PATH (when the program is invoked with no path)
 */
void
gethome (char **progpath)
{
	static char fnnam[] = "gethome";
	char *pp;
	static char *envpath, *path, *prog, *pathprog;
	int proglen, homdirlen;
	struct stat fbuf;
	char dircur[MAXPATH+1];
	int found;

	/* note that for error logging, we can't invoke logreturn() or
	   logexit(), because we don't have the path to the log file yet!
	 */

	homdir = prog = mystrdup(*progpath, fnnam);
	proglen = strlen(prog);
	if ((pp = strrchr(homdir, '/'))) {
		prgnam = mystrdup(pp+1, fnnam);
		*pp = '\0';
		if (! strcmp(homdir, ".")) {	/* prog invoked as ./<prog> */
			if (getcwd(dircur, MAXPATH) == NULL) {
				logexit(OUTPUT, LOG_ERR, ERRID_CWD_INDETERMINATE, __FILE__, __LINE__, fnnam, FALSE, "getcwd() failure, cannot determine current working directory");
			}
			free(homdir);
			homdir = mystrdup(dircur, fnnam);
		}
	} else {
		prgnam = mystrdup(prog, fnnam);
		/* there is no explicit path, so we have to get it from
	           the PATH environment variable
	         */
		if (! (pp = getenv("PATH"))) {
			logexit(OUTPUT, LOG_ERR, ERRID_ENV_PATH_INDETERMINATE, __FILE__, __LINE__, fnnam, FALSE, "cannot determine environment PATH");
		} else
			envpath = mystrdup(pp, fnnam);
		/* traverse the PATH, until we find PIKT home */
		if (strtok(envpath, ": ") == NULL) {
			logexit(OUTPUT, LOG_ERR, ERRID_ENV_PATH_INDETERMINATE, __FILE__, __LINE__, fnnam, FALSE, "cannot tokenize environment PATH");
		} else
			path = envpath;
		do {
			pathprog = (char *) mymalloc((strlen(path)+2+proglen) * sizeof(char), fnnam);
			sprintf(pathprog, "%s/%s", path, prog);
			found = ! lstat(pathprog, &fbuf);
			free(pathprog);
			if (found) {
				free(prog);
				homdir = mystrdup(path, fnnam);
				break;
			}
		} while ((path = strtok(NULL, ": ")));
		if (! path) {
			/* try again to determine homdir in getconf() */
			free(homdir);
			homdir = mystrdup(nullstring, fnnam);
		}
		free(envpath);
	}
	homdirlen = strlen(homdir);
	if (homdirlen) {
		if ((homdirlen >= 4) && ! strcmp(homdir+homdirlen-4, "/bin"))
			*(homdir+homdirlen-4) = '\0';
		else {
			free(homdir);
			homdir = mystrdup(nullstring, fnnam);
		}
	}
	return;
}

/* get PIKT.conf file settings
 */
void
getconf (char *hom)
{
	static char fnnam[] = "getconf";
	const char *testbranch;

	if (istest)
		testbranch = "/src/test";
	else
		testbranch = nullstring;
	/* look for PIKT.conf in $PIKTDIR/etc, then in /etc, then in
	   /etc/pikt, then in /usr/local/etc, finally in /usr/local/etc/pikt;
	   give up if in none of those locations
	 */
	PIKTconf = (char *) mymalloc((strlen(hom)+strlen(testbranch)+15) * sizeof(char), fnnam);
	sprintf(PIKTconf, "%s%s/etc/PIKT.conf", hom, testbranch);
	if ((confin = fopen(PIKTconf, "r")) == NULL) {
		free(PIKTconf);
		PIKTconf = mystrdup("/etc/PIKT.conf", fnnam);
		if ((confin = fopen(PIKTconf, "r")) == NULL) {
			free(PIKTconf);
			PIKTconf = mystrdup("/etc/pikt/PIKT.conf", fnnam);
			if ((confin = fopen(PIKTconf, "r")) == NULL) {
				free(PIKTconf);
				PIKTconf = mystrdup("/usr/local/etc/PIKT.conf", fnnam);
				if ((confin = fopen(PIKTconf, "r")) == NULL) {
					free(PIKTconf);
					PIKTconf = mystrdup("/usr/local/etc/pikt/PIKT.conf", fnnam);
					if ((confin = fopen(PIKTconf, "r")) == NULL) {
						logexit(OUTPUT, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, FALSE, "fopen() error for (possibly missing) PIKT.conf file");
					}
				}
			}
		}
	}
	lineno = 1;
	confparse();
	if (fclose(confin)) {
		logreturn(OUTPUT, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fclose() error for %s", PIKTconf);
		/* bailout = TRUE; */
	}
	if (bailout)
		return;
	/* now check for the presence of certain essential elements */
	if (! homdir || ! homdir[0]) {
		logreturn(OUTPUT, LOG_ERR, ERRID_PIKT_HOMDIR_INDETERMINATE, __FILE__, __LINE__, fnnam, FALSE, "unable to determine PIKT homdir, whether from the program path, the PATH environment variable, or from PIKT.conf");
		bailout = TRUE;
	}
	if (! master) {
		logreturn(OUTPUT, LOG_ERR, ERRID_PIKT_MASTER_INDETERMINATE, __FILE__, __LINE__, fnnam, FALSE, "in PIKT.conf, failure to specify master");
		bailout = TRUE;
	}
	if (master && ! strchr(master, '.')) {
		logreturn(OUTPUT, LOG_ERR, ERRID_PIKT_MASTER_NOT_FQDN, __FILE__, __LINE__, fnnam, FALSE, "in PIKT.conf, master is not fully qualified network name");
		bailout = TRUE;
	}
	if (! domain) {
		logreturn(OUTPUT, LOG_ERR, ERRID_DOMAIN_INDETERMINATE, __FILE__, __LINE__, fnnam, FALSE, "in PIKT.conf, failure to specify domain");
		bailout = TRUE;
	}
	if (! accessmode) {
		logreturn(OUTPUT, LOG_ERR, ERRID_ACCESS_INDETERMINATE, __FILE__, __LINE__, fnnam, FALSE, "in PIKT.conf, failure to specify access");
		bailout = TRUE;
	}
	if (isauth_by_master_address && ! master_address) {
		logreturn(OUTPUT, LOG_ERR, ERRID_PIKT_MASTER_ADDRESS_INDETERMINATE, __FILE__, __LINE__, fnnam, FALSE, "in PIKT.conf, failure to specify master address");
		bailout = TRUE;
	}
	if (! isauth_by_master && ! isauth_by_master_address) {
		logreturn(OUTPUT, LOG_ERR, ERRID_PIKT_MASTER_AUTHENTICATE_INDETERMINATE, __FILE__, __LINE__, fnnam, FALSE, "in PIKT.conf, authenticate_by_master and/or authenticate_by_master_address\nmust be YES (TRUE or ON)");
		bailout = TRUE;
	}
	if (isauth_by_master_ethernet_address && ! master_ethernet_address) {
		logreturn(OUTPUT, LOG_ERR, ERRID_PIKT_MASTER_ETHERNET_ADDRESS_INDETERMINATE, __FILE__, __LINE__, fnnam, FALSE, "in PIKT.conf, failure to specify master ethernet address");
		bailout = TRUE;
	}
	if (data_encryption_type > MAX_ENCRYPTION_TYPE) {
		logreturn(OUTPUT, LOG_ERR, ERRID_DATA_ENCRYPTION_TYPE_INVALID, __FILE__, __LINE__, fnnam, FALSE, "in PIKT.conf, data_encryption_type (%d) exceeds max encryption_type (%d)", data_encryption_type, MAX_ENCRYPTION_TYPE);
		bailout = TRUE;
	}
	if (auth_encryption_type > MAX_ENCRYPTION_TYPE) {
		logreturn(OUTPUT, LOG_ERR, ERRID_AUTH_ENCRYPTION_TYPE_INVALID, __FILE__, __LINE__, fnnam, FALSE, "in PIKT.conf, auth_encryption_type (%d) exceeds max encryption_type (%d)", auth_encryption_type, MAX_ENCRYPTION_TYPE);
		bailout = TRUE;
	}
	if (! pvtkey) {
		pvtkey = (char *) mymalloc((strlen(PVTKEY)+1) * sizeof(char), fnnam);
		sprintf(pvtkey, PVTKEY);
	}
	if ((pvtkeylen = strlen(pvtkey)) > MAXKEYBYTES) {
		logreturn(OUTPUT, LOG_ERR, ERRID_PVTKEY_LEN_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "in PIKT.conf, pvtkeylen (%d) exceeds maximum (%d), specify another pvtkey", pvtkeylen, MAXKEYBYTES);
		bailout = TRUE;
	}
	/* note: we should add checks here to see if the pvtkey is "weak"
	   (other than in the blowfish sense following), for example, is
	   too short, is a dictionary word, etc., etc.
	 */
	/*
	if (bf_setkey(&bc, pvtkey, pvtkeylen) == ERROR) {
		logreturn(OUTPUT, LOG_ERR, ERRID_PVTKEY_WEAK, __FILE__, __LINE__, fnnam, FALSE, "in PIKT.conf, pvtkey %s is a \"weak\" key, specify another", pvtkey);
		bailout = TRUE;
	}
	 */
	return;	
}

/* from PIKT.conf, set the string parameter
 */
void
setstr (char *name, char **parameter, char *setting)
{
	*parameter = mystrdup(setting, "setstr");
	return;
}

/* from PIKT.conf, set/unset the specified option flag, in the case of
   boolean options, else set the numerical parameter
 */
void
setint (char *name, int *parameter, int setting)
{
	*parameter = setting;
	return;
}

/* from PIKT.conf, set/unset the specified bit in services, thereby
   enabling/disabling the specified piktc_svc service
 */
void
setsrv (char *name, unsigned long service, int setting)
{
	static char fnnam[] = "setsrv";

	if (setting == TRUE)
		services |= service;
	else if (setting == FALSE)
		services &= ~service;
	else {
		logreturn(OUTPUT, LOG_ERR, ERRID_SERVICE_SETTING_INVALID, __FILE__, __LINE__, fnnam, FALSE, "for service %s, invalid setting %d", name, setting);
	}
	return;
}

/* initialize dir/file paths
 */
void
initpaths (char *hom)
{
	static char fnnam[] = "initpaths";
	const char *testbranch;

	if (istest)
		testbranch = "/src/test";
	else
		testbranch = nullstring;

	/* there are lots of redundant calls to strlen() here; we could
	   save strlen() return values to local vars, but in the interest
	   of simplicity we choose not to
	 */

	if (! bindir) {
		bindir = (char *) mymalloc((strlen(hom)+5) * sizeof(char), fnnam);
		sprintf(bindir, "%s/bin", hom);	/* we don't insert
						   testbranch into bindir
						 */
	}

	if (! etcdir) {
		etcdir = (char *) mymalloc((strlen(hom)+strlen(testbranch)+5) * sizeof(char), fnnam);
		sprintf(etcdir, "%s%s/etc", hom, testbranch);
	}

	if (! lokdir) {
		lokdir = (char *) mymalloc((strlen(hom)+strlen(testbranch)+5) * sizeof(char), fnnam);
		sprintf(lokdir, "%s%s/etc", hom, testbranch);
	}

	if (! libdir) {
		libdir = (char *) mymalloc((strlen(hom)+strlen(testbranch)+5) * sizeof(char), fnnam);
		sprintf(libdir, "%s%s/lib", hom, testbranch);
	}

	if (! vardir) {
		vardir = (char *) mymalloc((strlen(hom)+strlen(testbranch)+5) * sizeof(char), fnnam);
		sprintf(vardir, "%s%s/var", hom, testbranch);
	}

	if (! altdir) {
		altdir = (char *) mymalloc((strlen(libdir)+8) * sizeof(char), fnnam);
		sprintf(altdir, "%s/alerts", libdir);
	}

	/* prgdir is specially set in pikt.c main() */
	if (! prgdir || ! *prgdir) {
		prgdir = (char *) mymalloc((strlen(libdir)+10) * sizeof(char), fnnam);
		sprintf(prgdir, "%s/programs", libdir);
	}

	if (! objdir) {
		objdir = (char *) mymalloc((strlen(libdir)+9) * sizeof(char), fnnam);
		sprintf(objdir, "%s/objects", libdir);
	}

	if (! cfgdir) {
		cfgdir = (char *) mymalloc((strlen(libdir)+9) * sizeof(char), fnnam);
		sprintf(cfgdir, "%s/configs", libdir);
	}

	/* note: for now, one cannot override difdir or stgdir in PIKT.conf--
	   they are always assumed to be direct subdirectories off cfgdir
	 */

	if (! difdir) {
		difdir = (char *) mymalloc((strlen(cfgdir)+9) * sizeof(char), fnnam);
		sprintf(difdir, "%s/diffing", cfgdir);
	}

	if (! stgdir) {
		stgdir = (char *) mymalloc((strlen(cfgdir)+9) * sizeof(char), fnnam);
		sprintf(stgdir, "%s/staging", cfgdir);
	}

	if (! hstdir) {
		hstdir = (char *) mymalloc((strlen(vardir)+11) * sizeof(char), fnnam);
		sprintf(hstdir, "%s/histories", vardir);
	}

	if (! logdir) {
		logdir = (char *) mymalloc((strlen(vardir)+5) * sizeof(char), fnnam);
		sprintf(logdir, "%s/log", vardir);
	}

	/* already determined in getconf()
	PIKTconf = (char *) mymalloc((strlen(etcdir)+11) * sizeof(char), fnnam);
	sprintf(PIKTconf, "%s/PIKT.conf", etcdir);
	 */

	keysconf = (char *) mymalloc((strlen(etcdir)+11) * sizeof(char), fnnam);
	sprintf(keysconf, "%s/keys.conf", etcdir);

	pikt = (char *) mymalloc((strlen(bindir)+6) * sizeof(char), fnnam);
	sprintf(pikt, "%s/pikt", bindir);

	piktd = (char *) mymalloc((strlen(bindir)+7) * sizeof(char), fnnam);
	sprintf(piktd, "%s/piktd", bindir);

	piktc = (char *) mymalloc((strlen(bindir)+7) * sizeof(char), fnnam);
	sprintf(piktc, "%s/piktc", bindir);

	piktc_svc = (char *) mymalloc((strlen(bindir)+11) * sizeof(char), fnnam);
	sprintf(piktc_svc, "%s/piktc_svc", bindir);

	pikts = (char *) mymalloc((strlen(bindir)+7) * sizeof(char), fnnam);
	sprintf(pikts, "%s/pikts", bindir);

	piktx = (char *) mymalloc((strlen(bindir)+7) * sizeof(char), fnnam);
	sprintf(piktx, "%s/piktx", bindir);

	piktdconf = (char *) mymalloc((strlen(etcdir)+12) * sizeof(char), fnnam);
	sprintf(piktdconf, "%s/piktd.conf", etcdir);

	piktdconftmp = (char *) mymalloc((strlen(piktdconf)+5) * sizeof(char), fnnam);
	sprintf(piktdconftmp, "%s.tmp", piktdconf);

	piktdlock = (char *) mymalloc((strlen(lokdir)+12) * sizeof(char), fnnam);
	sprintf(piktdlock, "%s/piktd.lock", lokdir);

	piktdlog = (char *) mymalloc((strlen(logdir)+11) * sizeof(char), fnnam);
	sprintf(piktdlog, "%s/piktd.log", logdir);

	piktdalt = (char *) mymalloc((strlen(etcdir)+8) * sizeof(char), fnnam);
	sprintf(piktdalt, "%s/alerts", etcdir);

	piktclock = (char *) mymalloc((strlen(lokdir)+12) * sizeof(char), fnnam);
	sprintf(piktclock, "%s/piktc.lock", lokdir);

	piktclog = (char *) mymalloc((strlen(logdir)+11) * sizeof(char), fnnam);
	sprintf(piktclog, "%s/piktc.log", logdir);

	piktc_svclock = (char *) mymalloc((strlen(lokdir)+16) * sizeof(char), fnnam);
	sprintf(piktc_svclock, "%s/piktc_svc.lock", lokdir);

	/* the last path set; see mymalloc(), mycalloc() & mystrdup() in mem.c */
	piktc_svclog = (char *) mymalloc((strlen(logdir)+15) * sizeof(char), fnnam);
	sprintf(piktc_svclog, "%s/piktc_svc.log", logdir);

	if (! arpcmd)
		arpcmd = mystrdup(ARPCMD, fnnam);

	if (! cksumcmd)
		cksumcmd = mystrdup(CKSUMCMD, fnnam);

	if (! cpcmd)
		cpcmd = mystrdup(CPCMD, fnnam);

	if (! diffcmd)
		diffcmd = mystrdup(DIFFCMD, fnnam);

	if (! echocmd)
		echocmd = mystrdup(ECHOCMD, fnnam);
	echocmdlen = strlen(echocmd);

	if (! egrepcmd)
		egrepcmd = mystrdup(EGREPCMD, fnnam);

	if (! filecmd)
		filecmd = mystrdup(FILECMD, fnnam);

	if (! pscmd)
		pscmd = mystrdup(PSCMD, fnnam);

	if (! rmcmd)
		rmcmd = mystrdup(RMCMD, fnnam);

	if (! nicecmd)
		nicecmd = mystrdup(NICECMD, fnnam);

	if (! catcmd)
		catcmd = mystrdup(CATCMD, fnnam);

	if (! rpcinfocmd)
		rpcinfocmd = mystrdup(RPCINFOCMD, fnnam);

	#ifdef SCPCMD
	if (! scpcmd)
		scpcmd = mystrdup(SCPCMD, fnnam);
	#endif

	if (! sumcmd)
		sumcmd = mystrdup(SUMCMD, fnnam);

	if (! tailcmd)
		tailcmd = mystrdup(TAILCMD, fnnam);

	if (! trcmd)
		trcmd = mystrdup(TRCMD, fnnam);

	if (! bakext)
		bakext = mystrdup(BAKEXT, fnnam);

	/* now check for the presence of certain essential filepaths */
	if (! isfile(homdir, FALSE)) {
		logreturn(OUTPUT, LOG_ERR, ERRID_FILE_NOT_FOUND, __FILE__, __LINE__, fnnam, FALSE, "homdir not found");
		bailout = TRUE;
	}
	if (! isfile(bindir, FALSE)) {
		logreturn(OUTPUT, LOG_ERR, ERRID_FILE_NOT_FOUND, __FILE__, __LINE__, fnnam, FALSE, "bindir not found");
		bailout = TRUE;
	}
	if (! isfile(etcdir, FALSE)) {
		logreturn(OUTPUT, LOG_ERR, ERRID_FILE_NOT_FOUND, __FILE__, __LINE__, fnnam, FALSE, "etcdir not found");
		bailout = TRUE;
	}
	if (! isfile(lokdir, FALSE)) {
		logreturn(OUTPUT, LOG_ERR, ERRID_FILE_NOT_FOUND, __FILE__, __LINE__, fnnam, FALSE, "lokdir not found");
		bailout = TRUE;
	}
	if (! isfile(libdir, FALSE)) {
		logreturn(OUTPUT, LOG_ERR, ERRID_FILE_NOT_FOUND, __FILE__, __LINE__, fnnam, FALSE, "libdir not found");
		bailout = TRUE;
	}
	if (! isfile(vardir, FALSE)) {
		logreturn(OUTPUT, LOG_ERR, ERRID_FILE_NOT_FOUND, __FILE__, __LINE__, fnnam, FALSE, "vardir not found");
		bailout = TRUE;
	}
	return;
}

/* do the first and second given host names match?
   'vienna' matches 'vienna'
   'vienna' matches 'vienna.uppity.edu'
   'vienna.uppity.edu' matches 'vienna'
   'acme.com' matches 'acme.com.uk' and vice-versa--practically speaking,
   is this a problem?  we could check for this weird case, but at the cost
   of slowing down all match checks
 */
int
hostmatch (char *hn1, char *hn2)
{
	register int i;
	register char c1, c2;

	i = 0;
	do {
		c1 = *(hn1+i);
		c2 = *(hn2+i);
		if (! c1 || ! c2)
			break;
		if (c1 != c2)
			return FALSE;
		i++;
	} while (TRUE);
	if (! c1) {
		if (! c2 || (c2 == '.'))
			return TRUE;
	} else if (! c2) {
		if (! c1 || (c1 == '.'))
			return TRUE;
	}
	return FALSE;
}


/* when passed a dir arg such as ALTDIR from the piktmaster, find its (slave's)
   local equivalent; returns a mystrdup()'ed string that must be free()'ed
   later; note: for efficiency's sake, this is dependent on ALTDIR (and so on)
   being of the form "ALTDIR" (and so on), with the first char being uppercase;
   see the #define's in src/lib/common.h
 */
char *
getlocdir (char *argdir)
{
	static char fnnam[] = "getlocdir";
	char *locdir;

	locdir = (char *) NULL;
	if (isupper((int)*argdir)) {
		/* order these from most likely to least likely */
		if (! strcmp(argdir, ALTDIR))
			locdir = mystrdup(altdir, fnnam);
		else if (! strcmp(argdir, PRGDIR))
			locdir = mystrdup(prgdir, fnnam);
		else if (! strcmp(argdir, OBJDIR))
			locdir = mystrdup(objdir, fnnam);
		else if (! strcmp(argdir, HSTDIR))
			locdir = mystrdup(hstdir, fnnam);
		else if (! strcmp(argdir, LOGDIR))
			locdir = mystrdup(logdir, fnnam);
		else if (! strcmp(argdir, ETCDIR))
			locdir = mystrdup(etcdir, fnnam);
		else if (! strcmp(argdir, BINDIR))
			locdir = mystrdup(bindir, fnnam);
		else if (! strcmp(argdir, LOKDIR))
			locdir = mystrdup(lokdir, fnnam);
		else if (! strcmp(argdir, VARDIR))
			locdir = mystrdup(vardir, fnnam);
		else if (! strcmp(argdir, LIBDIR))
			locdir = mystrdup(libdir, fnnam);
		else if (! strcmp(argdir, DIFDIR))
			locdir = mystrdup(difdir, fnnam);
		else if (! strcmp(argdir, STGDIR))
			locdir = mystrdup(stgdir, fnnam);
	} else {
		locdir = mystrdup(argdir, fnnam);
	}
	return locdir;
}

/* return TRUE if file/dir exists, FALSE if not
 */
int
isfile (char *file, int rpterr)
{
	static char fnnam[] = "isfile";
	struct stat fbuf;

	if (stat(file, &fbuf)) {
		if (rpterr) {
			if (logdir) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "stat() error for %s", file);
			} else {
				logreturn(OUTPUT, LOG_ERR, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "stat() error for %s", file);
			}
		}
		return FALSE;
	} else
		return TRUE;
}

/* if we have an etcdir/private_key file overriding any PIKT.conf private_key
   or compiled-in private_key, get it
 */
void
getkey (char **key, char *keynam, char *dir)
{
	static char fnnam[] = "getkey";
	char *keypath;
	struct stat fbuf;
	FILE *keyfile;
	char *cp;

	keypath = (char *) mymalloc((strlen(etcdir)+strlen(keynam)+2)*sizeof(char), fnnam);
	sprintf(keypath, "%s/%s", etcdir, keynam);
	if (isfile(keypath, FALSE)) {
		if (stat(keypath, &fbuf) != 0) {
			logreturn(PIKTLOG, LOG_WARNING, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "stat() error for %s", keypath);
			bailout = TRUE;
		/* +1 to allow for possible file-ending '\n' */
		} else if (fbuf.st_size > MAXKEYBYTES+1) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_PVTKEY_LEN_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, TRUE, "%s keysize exceeds maximum (%d)", keypath, MAXKEYBYTES);
			bailout = TRUE;
		} else if ((keyfile = fopen(keypath, "r")) == NULL) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", keypath);
			bailout = TRUE;
		} else {
			FREE(*key);
			/* +2 for possible '\n' and string-ending '\0' */
			*key = (signed char *) mycalloc(MAXKEYBYTES+2, sizeof(signed char), fnnam);
			/* +1 to allow for possible file-ending '\n' */
			if (fread(*key, sizeof(signed char), MAXKEYBYTES+1, keyfile) == 0) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_READ_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fread() error for %s", keypath);
				bailout = TRUE;
			} else {
				/* dispose of '\n' if present */
				if ((cp = strrchr(*key, '\n')))
					*cp = '\0';
				/* ensure that *key is no more than
				   MAXKEYBYTES chars long
				 */
				*(*key+MAXKEYBYTES) = '\0';
			}
		}
	}
	free(keypath);
	return;
}

/* set up in daemon mode
 */
int
daemon_init (const int closefd)
{
	static char fnnam[] = "daemon_init";
	pid_t pid;
	struct rlimit rl;
	int fdmax = 0;
	int i;
	int fd;

	if ((pid = fork()) < 0) {
		logexit(PIKTLOG, LOG_ERR, ERRID_FORK_ERROR, __FILE__, __LINE__, fnnam, FALSE, "fork() error");
	} else if (pid != 0) {
		exit(0);	/* parent goes bye-bye */
	}

	/* child continues */

	if (closefd) {
		/* determine file descriptor resource limits */
		#ifdef RLIMIT_NOFILE
		if (getrlimit(RLIMIT_NOFILE, &rl) < 0) {
			logexit(PIKTLOG, LOG_ERR, ERRID_GETRLIMIT_ERROR, __FILE__, __LINE__, fnnam, FALSE, "getrlimit() error, can't determine rlim_max");
		} else {
			fdmax = rl.rlim_max;
		}
		#else
		fdmax = getdtablesize();
		#endif
		/* close any open file descriptors */
		for (i=0; i<fdmax; i++)
			close(i);
		/* redirect stdin, stdout & stderr to /dev/null */
		fd = open("/dev/null", O_RDWR);
		(void)dup2(fd, STDIN_FILENO);
		(void)dup2(fd, STDOUT_FILENO);
		(void)dup2(fd, STDERR_FILENO);
	}

	/* become session leader */
	#ifdef HAVE_SETSID
	setsid();
	#endif

	chdir("/");		/* change working directory */

	umask(0);		/* clear our file mode creation mask */

	return(0);
}
