%{

/*
   cmd.y -- PIKT piktc command-line parser

   $Header: /var/cvs/PIKT/pikt/src/piktc/cmd.y,v 1.23 2004/06/19 23:12:43 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002,2003,2004 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#include "../lib/common.h"
#include "piktc.h"

#define SETBS(a,b)	tbs1 = a; \
			a = b; \
			free(tbs1)

unsigned long *tbs1, *tbs2, *tbs3;

%}

%union {
	signed char *sval;
}

%token <sval> _DEFINE _ALERT _PROGRAM _MYFILE _OBJECT _HOST _COMMAND
%token _MINUS _PLUS _OPT _DEFOPT _ALTOPT _PRGOPT _FILOPT _OBJOPT _HSTOPT _ALL _CMDOPT

%%

start:		/* empty */ {
			if (scanpass == 1) {
				ishelp = TRUE;	/* never reached in 2nd pass */
			}
		}
		| options {
			if (scanpass == 1) {
				bailout = TRUE;	/* never reached in 2nd pass */
			}
		}
		| options hostitems
		| options items hostitems
		;

options:	option
		| options option
		;

option:		_MINUS opts
		/* | _PLUS opts	*/	/* no plus opts now, maybe in future */
		;

opts:		_OPT
		| opts _OPT
		;

items:		item
		| items item
		;

item:		_PLUS initems
		| _MINUS exitems
		| _PLUS _CMDOPT commands
		;

hostitems:	hostitem
		| hostitems hostitem
		;

hostitem:	_PLUS inhostitems
		| _MINUS exhostitems
		;

initems:	_DEFOPT indefines
		| _DEFOPT {
			/* no defines */
			if (scanpass == 1)
				bailout = TRUE;	/* never reached in 2nd pass */
		}
		| _ALTOPT _ALL {
			if (scanpass == 1) {
				regname(&incmdalerts, "all", FALSE);
				alertsall = TRUE;
			}
		}
		| _ALTOPT inalerts
		| _ALTOPT {
			/* no alerts */
			if (scanpass == 1)
				bailout = TRUE; /* never reached in 2nd pass */
		}
		| _PRGOPT _ALL {
			if (scanpass == 1 ) {
				regname(&incmdprograms, "all", FALSE);
				programsall = TRUE;
			}
		}
		| _PRGOPT inprograms
		| _PRGOPT {
			/* no programs */
			if (scanpass == 1)
				bailout = TRUE;	/* never reached in 2nd pass */
		}
		| _FILOPT _ALL {
			if (scanpass == 1 ) {
				regname(&incmdfiles, "all", FALSE);
				filesall = TRUE;
			}
		}
		| _FILOPT infiles
		| _FILOPT {
			/* no files */
			if (scanpass == 1)
				bailout = TRUE;	/* never reached in 2nd pass */
		}
		| _OBJOPT _ALL {
			if (scanpass == 1) {
				regname(&incmdobjects, "all", FALSE);
				objectsall = TRUE;
			}
		}
		| _OBJOPT inobjects
		| _OBJOPT {
			/* no objects */
			if (scanpass == 1)
				bailout = TRUE;	/* never reached in 2nd pass */
		}
		;

inhostitems:	_HSTOPT _ALL {
			if (scanpass == 1) {
				regname(&incmdhosts, "all", FALSE);
				hostsall = TRUE;
			}
		}
		| _HSTOPT inhosts
		| _HSTOPT {
			/* no hosts */
			if (scanpass == 1)
				bailout = TRUE;	/* never reached in 2nd pass */
		}
		;

indefines:	indefine
		| indefines indefine
		;

indefine:	_DEFINE {
			if (scanpass == 2) {
				tbs3 = getbs($1, DEF, FALSE);
				SETBS(gdbs, bsunion(gdbs, tbs3, dbslongs));
				tbs2 = bscomplement(tbs3, dbslongs);
				SETBS(gubs, bsintersection(gubs, tbs2, dbslongs));
				free(tbs2);
			}
			free($1);
		}
		;

inalerts:	_ALERT {
			if (scanpass == 1)
				regname(&incmdalerts, $1, FALSE);
			free($1);
		}
		| inalerts _ALERT {
			if (scanpass == 1)
				regname(&incmdalerts, $2, FALSE);
			free($2);
		}
		;

inprograms:	_PROGRAM {
			if (scanpass == 1)
				regname(&incmdprograms, $1, FALSE);
			free($1);
		}
		| inprograms _PROGRAM {
			if (scanpass == 1)
				regname(&incmdprograms, $2, FALSE);
			free($2);
		}
		;

infiles:	_MYFILE {
			if (scanpass == 1)
				regname(&incmdfiles, $1, FALSE);
			free($1);
		}
		| infiles _MYFILE {
			if (scanpass == 1)
				regname(&incmdfiles, $2, FALSE);
			free($2);
		}
		;

inobjects:	_OBJECT {
			if (scanpass == 1)
				regname(&incmdobjects, $1, FALSE);
			free($1);
		}
		| inobjects _OBJECT {
			if (scanpass == 1)
				regname(&incmdobjects, $2, FALSE);
			free($2);
		}
		;

inhosts:	_HOST {
			if (scanpass == 1)
				regname(&incmdhosts, $1, FALSE);
			free($1);
		}
		| inhosts _HOST {
			if (scanpass == 1)
				regname(&incmdhosts, $2, FALSE);
			free($2);
		}
		;

exitems:	_DEFOPT exdefines
		| _DEFOPT {
			/* no defines */
			if (scanpass == 1)
				bailout = TRUE;	/* never reached in 2nd pass */
		}
		| _ALTOPT _ALL {
			if (scanpass == 1) {
				regname(&excmdalerts, "all", FALSE);
				alertsnone = TRUE;
			}
		}
		| _ALTOPT exalerts
		| _ALTOPT {
			/* no alerts */
			if (scanpass == 1)
				bailout = TRUE;	/* never reached in 2nd pass */
		}
		| _PRGOPT _ALL {
			if (scanpass == 1) {
				regname(&excmdprograms, "all", FALSE);
				programsnone = TRUE;
			}
		}
		| _PRGOPT exprograms
		| _PRGOPT {
			/* no programs */
			if (scanpass == 1)
				bailout = TRUE;	/* never reached in 2nd pass */
		}
		| _FILOPT _ALL {
			if (scanpass == 1) {
				regname(&excmdfiles, "all", FALSE);
				filesnone = TRUE;
			}
		}
		| _FILOPT exfiles
		| _FILOPT {
			/* no files */
			if (scanpass == 1)
				bailout = TRUE;	/* never reached in 2nd pass */
		}
		| _OBJOPT _ALL {
			if (scanpass == 1) {
				regname(&excmdobjects, "all", FALSE);
				objectsnone = TRUE;
			}
		}
		| _OBJOPT exobjects
		| _OBJOPT {
			/* no objects */
			if (scanpass == 1)
				bailout = TRUE;	/* never reached in 2nd pass */
		}
		;

exhostitems:	_HSTOPT _ALL {
			if (scanpass == 1) {
				regname(&excmdhosts, "all", FALSE);
				hostsnone = TRUE;
			}
		}
		| _HSTOPT exhosts
		| _HSTOPT {
			/* no hosts */
			if (scanpass == 1)
				bailout = TRUE;	/* never reached in 2nd pass */
		}
		;

exdefines:	exdefine
		| exdefines exdefine
		;

exdefine:	_DEFINE {
			if (scanpass == 2) {
				tbs3 = getbs($1, DEF, FALSE);
				SETBS(gubs, bsunion(gubs, tbs3, dbslongs));
				tbs2 = bscomplement(tbs3, dbslongs);
				SETBS(gdbs, bsintersection(gdbs, tbs2, dbslongs));
				free(tbs2);
			}
			free($1);
		}
		;

exalerts:	_ALERT {
			if (scanpass == 1)
				regname(&excmdalerts, $1, FALSE);
			free($1);
		}
		| exalerts _ALERT {
			if (scanpass == 1)
				regname(&excmdalerts, $2, FALSE);
			free($2);
		}
		;

exprograms:	_PROGRAM {
			if (scanpass == 1)
				regname(&excmdprograms, $1, FALSE);
			free($1);
		}
		| exprograms _PROGRAM {
			if (scanpass == 1)
				regname(&excmdprograms, $2, FALSE);
			free($2);
		}
		;

exfiles:	_MYFILE {
			if (scanpass == 1)
				regname(&excmdfiles, $1, FALSE);
			free($1);
		}
		| exfiles _MYFILE {
			if (scanpass == 1)
				regname(&excmdfiles, $2, FALSE);
			free($2);
		}
		;

exobjects:	_OBJECT {
			if (scanpass == 1)
				regname(&excmdobjects, $1, FALSE);
			free($1);
		}
		| exobjects _OBJECT {
			if (scanpass == 1)
				regname(&excmdobjects, $2, FALSE);
			free($2);
		}
		;

exhosts:	_HOST {
			if (scanpass == 1)
				regname(&excmdhosts, $1, FALSE);
			free($1);
		}
		| exhosts _HOST {
			if (scanpass == 1)
				regname(&excmdhosts, $2, FALSE);
			free($2);
		}
		;

commands:	command
		| commands command
		;

command:	_COMMAND {
			if (scanpass == 1)
				regname(&incmdcommands, $1, TRUE);
			free($1);
		}
		;

%%
